/*************************************************************************
 * ArtisticStylePlugin: Using Artistic Style at Qt Creator.              *
 * Copyright (C) 2010-2011  Lorenz Haas <lykurg@gmail.com>               *
 *                                                                       *
 * This program is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *************************************************************************/

#include "artisticstyleplugin.h"
#include "artisticstyleiterator.h"
#include "artisticstyleoptionpage.h"
#include "artisticstyleconstants.h"

#include "dtl-cpp/dtl.hpp"

#include <coreplugin/icore.h>
#include <coreplugin/coreconstants.h>
#include <coreplugin/editormanager/ieditor.h>
#include <coreplugin/editormanager/editormanager.h>
#include <coreplugin/actionmanager/actioncontainer.h>
#include <coreplugin/actionmanager/actionmanager.h>
#include <coreplugin/actionmanager/command.h>

#include <texteditor/basetextdocumentlayout.h>

#include <QPlainTextEdit>
#include <QSettings>
#include <QTextStream>
#include <QtPlugin>
#include <QScrollBar>
#include <QTextBlock>
#include <QMenu>

namespace ArtisticStyle
{
  namespace Internal
  {
    ArtisticStylePlugin::ArtisticStylePlugin()
    {
    }

    ArtisticStylePlugin::~ArtisticStylePlugin()
    {
    }

    bool ArtisticStylePlugin::initialize(const QStringList& /*args*/, QString* /*errMsg*/)
    {
      Core::ActionManager* am = Core::ICore::instance()->actionManager();
      m_action = new QAction(this);
      m_action->setText("Format source code");
      connect(m_action, SIGNAL(triggered(bool)), this, SLOT(formatSource()));
      Core::Command* cmd;
#if (IDE_VERSION_MAJOR == 2 && IDE_VERSION_MINOR == 1)
      cmd = am->registerAction(m_action,
                               "ArtisticStylePlugin.FormatSourceCode",
                               Core::Context(Core::Constants::C_GLOBAL));
#else
      cmd = am->registerAction(m_action,
                               "ArtisticStylePlugin.FormatSourceCode",
                               QList<int>() << Core::Constants::C_GLOBAL_ID);
#endif



      cmd->setDefaultKeySequence(QKeySequence("Ctrl+Y"));
      am->actionContainer(Core::Constants::M_EDIT)->menu()->addAction(cmd->action());
      m_action->setEnabled(false);
      const Core::EditorManager* editorManager = Core::EditorManager::instance();

      if (editorManager)
        connect(editorManager, SIGNAL(currentEditorChanged(Core::IEditor*)),
                this, SLOT(updateAction(Core::IEditor*)));

      ArtisticStyleOptionPage* options = new ArtisticStyleOptionPage;
      addAutoReleasedObject(options);
      QObject::connect(options, SIGNAL(preferencesChanged()), SLOT(updateFormatter()));
      updateFormatter();
      return true;
    }

    void ArtisticStylePlugin::updateAction(Core::IEditor* editor)
    {
      m_action->setEnabled(!(!editor || !qobject_cast<QPlainTextEdit*>(editor->widget())));
      return;
    }

    void ArtisticStylePlugin::extensionsInitialized()
    {
    }

    void ArtisticStylePlugin::shutdown()
    {
    }

    void ArtisticStylePlugin::updateFormatter()
    {
      configureFormatter(m_formatter);
    }

    void ArtisticStylePlugin::configureFormatter(astyle::ASFormatter& formatter)
    {
      QSettings* settings = Core::ICore::instance()->settings();
      settings->beginGroup("ArtisticStylePlugin");

      // Predefined Style Options
      switch (settings->value(Constants::ASP_STYLE, Constants::ASP_STYLE_DEF).toInt())
      {
        case astyle::STYLE_ALLMAN:
          formatter.setFormattingStyle(astyle::STYLE_ALLMAN);
          break;
        case astyle::STYLE_JAVA:
          formatter.setFormattingStyle(astyle::STYLE_JAVA);
          break;
        case astyle::STYLE_KandR:
          formatter.setFormattingStyle(astyle::STYLE_KandR);
          break;
        case astyle::STYLE_STROUSTRUP:
          formatter.setFormattingStyle(astyle::STYLE_STROUSTRUP);
          break;
        case astyle::STYLE_WHITESMITH:
          formatter.setFormattingStyle(astyle::STYLE_WHITESMITH);
          break;
        case astyle::STYLE_BANNER:
          formatter.setFormattingStyle(astyle::STYLE_BANNER);
          break;
        case astyle::STYLE_GNU:
          formatter.setFormattingStyle(astyle::STYLE_GNU);
          break;
        case astyle::STYLE_LINUX:
          formatter.setFormattingStyle(astyle::STYLE_LINUX);
          break;
        case astyle::STYLE_HORSTMANN:
          formatter.setFormattingStyle(astyle::STYLE_HORSTMANN);
          break;
        case astyle::STYLE_1TBS:
          formatter.setFormattingStyle(astyle::STYLE_1TBS);
          break;
        default:
          formatter.setFormattingStyle(astyle::STYLE_NONE);
          break;
      }

      // Tab and Bracket Options
      if (settings->value(Constants::ASP_INDENT, Constants::ASP_INDENT_DEF).toBool())
      {
        QString style = settings->value(Constants::ASP_INDENTSTYLE, Constants::ASP_INDENTSTYLE_DEF).toString();
        int size = settings->value(Constants::ASP_INDENTSIZE, Constants::ASP_INDENTSIZE_DEF).toInt();
        if (size < 2 || size > 20)
          size = 2;

        if (style == "tab")
          formatter.setTabIndentation(size, false);
        else if (style == "force-tab")
          formatter.setTabIndentation(size, true);
        else
          formatter.setSpaceIndentation(size);
        formatter.setIndentManuallySet(true);
      }

      if (settings->value(Constants::ASP_BRACKETS, Constants::ASP_BRACKETS_DEF).toBool())
      {
        switch (settings->value(Constants::ASP_BRACKETSSTYLE, Constants::ASP_BRACKETSSTYLE_DEF).toInt())
        {
          case astyle::ATTACH_MODE:
            formatter.setBracketFormatMode(astyle::ATTACH_MODE);
            break;
          case astyle::LINUX_MODE:
            formatter.setBracketFormatMode(astyle::LINUX_MODE);
            break;
          case astyle::STROUSTRUP_MODE:
            formatter.setBracketFormatMode(astyle::STROUSTRUP_MODE);
            break;
          case astyle::HORSTMANN_MODE:
            formatter.setBracketFormatMode(astyle::HORSTMANN_MODE);
            break;
          case astyle::BREAK_MODE:
          default:
            formatter.setBracketFormatMode(astyle::BREAK_MODE);
            break;
        }
      }

      // Indentation Options
      if (settings->value(Constants::ASP_INDENTCLASSES, Constants::ASP_INDENTCLASSES_DEF).toBool())
        formatter.setClassIndent(true);
      if (settings->value(Constants::ASP_INDENTSWITCHES, Constants::ASP_INDENTSWITCHES_DEF).toBool())
        formatter.setSwitchIndent(true);
      if (settings->value(Constants::ASP_INDENTCASES, Constants::ASP_INDENTCASES_DEF).toBool())
        formatter.setCaseIndent(true);
      if (settings->value(Constants::ASP_INDENTBRACKETS, Constants::ASP_INDENTBRACKETS_DEF).toBool())
        formatter.setBracketIndent(true);
      if (settings->value(Constants::ASP_INDENTBLOCKS, Constants::ASP_INDENTBLOCKS_DEF).toBool())
        formatter.setBlockIndent(true);
      if (settings->value(Constants::ASP_INDENTNAMESPACES, Constants::ASP_INDENTNAMESPACES_DEF).toBool())
        formatter.setNamespaceIndent(true);
      if (settings->value(Constants::ASP_INDENTLABELS, Constants::ASP_INDENTLABELS_DEF).toBool())
        formatter.setLabelIndent(true);
      if (settings->value(Constants::ASP_INDENTPREPROCESSOR, Constants::ASP_INDENTPREPROCESSOR_DEF).toBool())
        formatter.setPreprocessorIndent(true);
      if (settings->value(Constants::ASP_INDENTCOL1COMMENTS, Constants::ASP_INDENTCOL1COMMENTS_DEF).toBool())
        formatter.setIndentCol1CommentsMode(true);
      if (settings->value(Constants::ASP_MAXINSTATEMENTINDENT, Constants::ASP_MAXINSTATEMENTINDENT_DEF).toBool())
      {
        int size = settings->value(Constants::ASP_MAXINSTATEMENTINDENTSIZE, Constants::ASP_MAXINSTATEMENTINDENTSIZE_DEF).toInt();
        if (size < 2 || size > 80)
          size = 40;
        formatter.setMaxInStatementIndentLength(size);
      }
      if (settings->value(Constants::ASP_MINCONDITIONALINDENT, Constants::ASP_MINCONDITIONALINDENT_DEF).toBool())
      {
        int size = settings->value(Constants::ASP_MINCONDITIONALINDENTSIZE, Constants::ASP_MINCONDITIONALINDENTSIZE_DEF).toInt();
        if (size < 2 || size > 40)
          size = 4;
        formatter.setMinConditionalIndentLength(size);
        formatter.setMinConditionalManuallySet(true);
      }

      // Padding Options
      if (settings->value(Constants::ASP_BREAKBLOCKSALL, Constants::ASP_BREAKBLOCKSALL_DEF).toBool())
      {
        formatter.setBreakBlocksMode(true);
        formatter.setBreakClosingHeaderBlocksMode(true);
      }
      if (settings->value(Constants::ASP_BREAKBLOCKS, Constants::ASP_BREAKBLOCKS_DEF).toBool())
        formatter.setBreakBlocksMode(true);
      if (settings->value(Constants::ASP_PADOPER, Constants::ASP_PADOPER_DEF).toBool())
        formatter.setOperatorPaddingMode(true);
      if (settings->value(Constants::ASP_PADPAREN, Constants::ASP_PADPAREN_DEF).toBool())
      {
        formatter.setParensOutsidePaddingMode(true);
        formatter.setParensInsidePaddingMode(true);
      }
      if (settings->value(Constants::ASP_PADPARENOUT, Constants::ASP_PADPARENOUT_DEF).toBool())
        formatter.setParensOutsidePaddingMode(true);
      if (settings->value(Constants::ASP_PADPARENIN, Constants::ASP_PADPARENIN_DEF).toBool())
        formatter.setParensInsidePaddingMode(true);
      if (settings->value(Constants::ASP_PADHEADER, Constants::ASP_PADHEADER_DEF).toBool())
        formatter.setParensHeaderPaddingMode(true);
      if (settings->value(Constants::ASP_UNPADPAREN, Constants::ASP_UNPADPAREN_DEF).toBool())
        formatter.setParensUnPaddingMode(true);
      if (settings->value(Constants::ASP_DELETEEMPTYLINES, Constants::ASP_DELETEEMPTYLINES_DEF).toBool())
        formatter.setDeleteEmptyLinesMode(true);
      if (settings->value(Constants::ASP_FILLEMPTYLINES, Constants::ASP_FILLEMPTYLINES_DEF).toBool())
        formatter.setEmptyLineFill(true);

      // Formatting Options
      if (settings->value(Constants::ASP_BREAKCLOSINGBRACKETS, Constants::ASP_BREAKCLOSINGBRACKETS_DEF).toBool())
        formatter.setBreakClosingHeaderBracketsMode(true);
      if (settings->value(Constants::ASP_BREAKELSEIFS, Constants::ASP_BREAKELSEIFS_DEF).toBool())
        formatter.setBreakElseIfsMode(true);
      if (settings->value(Constants::ASP_ADDBRACKETS, Constants::ASP_ADDBRACKETS_DEF).toBool())
        formatter.setAddBracketsMode(true);
      if (settings->value(Constants::ASP_ADDONELINEBRACKETS, Constants::ASP_ADDONELINEBRACKETS_DEF).toBool())
        formatter.setAddOneLineBracketsMode(true);
      if (settings->value(Constants::ASP_KEEPONELINEBLOCKS, Constants::ASP_KEEPONELINEBLOCKS_DEF).toBool())
        formatter.setBreakOneLineBlocksMode(false);
      if (settings->value(Constants::ASP_KEEPONELINESTATEMENTS, Constants::ASP_KEEPONELINESTATEMENTS_DEF).toBool())
        formatter.setSingleStatementsMode(false);
      if (settings->value(Constants::ASP_CONVERTTABS, Constants::ASP_CONVERTTABS).toBool())
        formatter.setTabSpaceConversionMode(true);
      if (settings->value(Constants::ASP_ALIGNPOINTER, Constants::ASP_ALIGNPOINTER_DEF).toBool())
      {
        switch (settings->value(Constants::ASP_ALIGNPOINTERSTYLE, Constants::ASP_ALIGNPOINTERSTYLE_DEF).toInt())
        {
          case astyle::ALIGN_MIDDLE:
            formatter.setPointerAlignment(astyle::ALIGN_MIDDLE);
            break;
          case astyle::ALIGN_NAME:
            formatter.setPointerAlignment(astyle::ALIGN_NAME);
            break;
          case astyle::ALIGN_TYPE:
          default:
            formatter.setPointerAlignment(astyle::ALIGN_TYPE);
            break;
        }
      }

      formatter.setCStyle();
      formatter.setModeManuallySet(true);
      settings->endGroup();
    }

    void ArtisticStylePlugin::formatSource()
    {
      const Core::EditorManager* editorManager = Core::EditorManager::instance();
      if (!editorManager)
        return;

      Core::IEditor* editor = editorManager->currentEditor();
      if (!editor)
        return;

      QPlainTextEdit* e = qobject_cast<QPlainTextEdit*>(editor->widget());
      if (!e)
        return;

      QString filename = editor->displayName();
      if (!(filename.endsWith(".cpp") || filename.endsWith(".c") || filename.endsWith(".h") || filename.endsWith(".hpp")))
        return;


      QString sourceData = e->toPlainText();
      QTextStream in(&sourceData);
      ArtisticStyleIterator it(&in);
      m_formatter.init(&it);

      QString formattedData;
      while (m_formatter.hasMoreLines())
        formattedData += QString::fromLocal8Bit(m_formatter.nextLine().data()) + "\n";
      if (sourceData == formattedData)
        return;



      // Save folded blocks
      QList<int> collapsedBlocks;
      QTextBlock block = e->document()->firstBlock();
      while (block.isValid())
      {
        if (!block.userData())
        {
          block = block.next();
          continue;
        }

        TextEditor::TextBlockUserData* userdata = static_cast<TextEditor::TextBlockUserData*>(block.userData());
#if (IDE_VERSION_MAJOR == 2 && IDE_VERSION_MINOR == 1)
        if (userdata->folded())
        {
          collapsedBlocks += block.blockNumber();
          TextEditor::BaseTextDocumentLayout::doFoldOrUnfold(block, true);
        }
#else
        if (userdata->collapsed())
        {
          int number = block.blockNumber();
          if (userdata->collapseMode() == TextEditor::TextBlockUserData::CollapseThis)
          {
            number--;
            if (block.previous().isValid())
              TextEditor::TextBlockUserData::doCollapse(block.previous(), true);
          }
          else
            TextEditor::TextBlockUserData::doCollapse(block, true);
          collapsedBlocks += number;
        }
#endif
        block = block.next();
      }
      e->update();



      // restore the cursor
      QString tmp = sourceData.left(e->textCursor().position());
      tmp.remove(" ");
      tmp.remove("\t");
      tmp.remove("\n");
      int cursorPos = tmp.size();
      int newCursorPos = -1;
      int indexSourceX = cursorPos;
      while (0 > newCursorPos)
      {
        QString tmp = formattedData.left(indexSourceX);
        tmp.remove(" ");
        tmp.remove("\t");
        tmp.remove("\n");
        if (tmp.size() == cursorPos)
          newCursorPos = indexSourceX;
        else
          indexSourceX += cursorPos - tmp.size();
      }

      QFontMetrics fm(e->document()->defaultFont());
      int offset = e->cursorRect().y() / fm.height();

      QTextCursor c = e->textCursor();
      c.beginEditBlock();

      // ---
      // dtl
      // ---
      dtl::Diff<char, std::string> diff(sourceData.toStdString(), formattedData.toStdString());
      diff.compose();
      std::vector< std::pair< char, dtl::elemInfo > > ses = diff.getSes().getSequence();
      std::vector< std::pair< char, dtl::elemInfo > >::iterator itNum = ses.begin();

      c.movePosition(QTextCursor::Start, QTextCursor::MoveAnchor);
      for (; itNum < ses.end(); itNum++)
      {
        switch ((*itNum).second.type)
        {
          case dtl::SES_ADD:
            {
              QString str = QString((*itNum).first);
              if (str == QLatin1String("\n"))
              {
                int number = c.blockNumber();
                for (int i = 0; i < collapsedBlocks.size(); ++i)
                  if (collapsedBlocks.at(i) > number)
                    ++collapsedBlocks[i];
              }
              c.insertText(str);
            }
            break;
          case dtl::SES_DELETE:
            {
              if (QString((*itNum).first) == QLatin1String("\n"))
              {
                int number = c.blockNumber();
                for (int i = 0; i < collapsedBlocks.size(); ++i)
                  if (collapsedBlocks.at(i) > number)
                    --collapsedBlocks[i];
              }
              c.deleteChar();
            }
            break;
          case dtl::SES_COMMON:
            c.movePosition(QTextCursor::Right, QTextCursor::MoveAnchor);
            break;
        }
      }

      c.endEditBlock();
      c.setPosition(newCursorPos);

      e->setTextCursor(c);


      // restore folded blocks
      QTextDocument* doc = e->document();
      foreach(int blockNumber, collapsedBlocks)
      {
        QTextBlock block = doc->findBlockByNumber(qMax(0, blockNumber));
        if (block.isValid())
#if (IDE_VERSION_MAJOR == 2 && IDE_VERSION_MINOR == 1)
          TextEditor::BaseTextDocumentLayout::doFoldOrUnfold(block, false);
#else
          TextEditor::TextBlockUserData::doCollapse(block, false);
#endif
      }


      offset = e->cursorRect().y() / fm.height() - offset;
      e->verticalScrollBar()->setValue(e->verticalScrollBar()->value() + offset);
      e->document()->setModified(true);
    }
  } // namespace Internal
} // namespace ArtisticStyle

Q_EXPORT_PLUGIN(ArtisticStyle::Internal::ArtisticStylePlugin)
